import time
import keysightSD1
import keysight_hvi as kthvi

"""
PathWave Test Sync Executive Integration with PathWave FPGA

Copyright Keysight Technologies 2022

The goal of this application example is to show how to use HVI Port blocks placed in the sandbox of an M3xxxA module FPGA to allow communication 
between an HVI sequence designed using HVI API and custom blocks loaded into an instrument FPGA sandbox using Pathwave FPGA.

Please note that there are two types of HVI Port blocks that can be placed within the FPGA sandbox of an instrument: HVI Port Memory Maps and 
HVI Port Register Banks. Read/write of both type of HVI Port blocks is tested in this application example.

The complete list of HVI API functionality showcased in this application example is as follows:
 * Read/write data from/to an HVI sequence to/from an HVI port (Memory Map) inserted in a custom FPGA FW
 * Read/write data from/to an HVI sequence to/from an HVI port (Register bank) inserted in a custom FPGA FW
 * Read/write PXI line values from a custom FPGA FW
 * Usage of HVI Actions and Events to communicate with a custom FPGA FW

For further info please check the programming example document titled "PathWave Test Sync Executive Integration with PathWave FPGA".
"""

######################################
# Application Configuration 
######################################

class ApplicationConfig:
    """
    Defines module descriptors, configuration options and names of HVI engines, actions, triggers
    """
    def __init__(self):
        """
        Define PXI instruments options
        """
        # Define options to open the instruments. 
        # Complete option list can be found in the SD1 3.x User Guide
        self.options = 'channelNumbering=keysight'
        # Simulated HW Option
        self.hardware_simulated = True
        """ 
        Define names of HVI engines, actions, events, triggers, registers
        """
        # HVI engine names
        self.leader_engine = "LeaderEngine"
        self.follower_engine = "FollowerEngine"
        # HVI register names
        self.hvi_quit = "HVI Quit"
        self.action4_cnt = "Action4 Counter"
        self.counter_reg = "Loop Counter"
        self.mem_map = "Memory Map Value"
        self.mem_map_counter = "Memory Map Counter"
        self.pxi_values = "PXI Values"
        self.follower_pxi_values = "Follower PXI Values"
        self.follower_counter_reg = "Follower Counter"
        # HVI events and actions
        self.hvi_user_event_4 =  "FpgaUserEvent4"
        self.hvi_user_action_4 = "FpgaUserAction4"
        """
        Define names of FPGA sandbox resources
        """
        # Bitstream files generated by compiling PathWave FPGA project files
        self.leader_project_file = "../Bitfiles/HviPortExampleLeader.k7z"
        self.follower_project_file = "../Bitfiles/HviPortExampleFollower.k7z"
        # Sandbox name defined by each instrument. See SD1 3.x User Guide for further info
        self.M3xxxA_sandbox = "sandbox0" # The M3xxxA_sandbox name is not arbitrary and cannot be changed 
        # FPGA Sandbox resource names
        # NOTE The FPGA resource names are not arbitrary. They correspond to the names defined in the PathWave FPGA project files
        self.num_leader_regs = 6 # number of mem. maps and registers placed in the leader PathWave FPGA project
        self.num_follower_regs = 3 # number of mem. maps and registers placed in the follower PathWave FPGA project
        self.memory_map = "MainEngine_Memory_1"
        self.reg_action4_cnt = "Register_Bank_HviAction4Cnt"
        self.reg_event4 = "Register_Bank_HviEvent4"
        self.reg_pxi_out = "Register_Bank_HviPxiTrigOut"
        self.reg_pxi_in = "Register_Bank_HviPxiTrigIn"
        self.follower_reg_pxi_in = "Register_Bank_HviPxiTrigIn"
        # Debug FPGA Sandbox Registers
        self.leader_debug_regs = [
            "Debug_HviAction4Counter",
            "Debug_HviEvent4Counter",
            "Debug_PxiTrigOut",
            "Debug_PxiTrigIn"]
        """
        Define HW Platform
        """
        # Define module descriptors below with your instruments information
        self.module_descriptors = [
            ModuleDescriptor('M3202A', 1, 4, self.options, self.leader_engine),
            ModuleDescriptor('M3202A', 1, 5, self.options, self.follower_engine)]
        
        # Define list of chassis containing instruments to be used by HVI
        self.chassis_list = [1]
        
        # Multi-chassis setup
        # Define the System Sync Modules included in your system.
        # For example: 
        # self.system_sync_module_descriptors = [
        #    SystemSyncModuleDescriptor('PXI0::CHASSIS2::SLOT10::INDEX0::INSTR', self.ssm_options)]
        self.ssm_options = ''
        self.ssm_simulation_options = 'Simulate=true,DriverSetup=Model=M9033A'
        self.system_sync_modules_descriptors = []
        # For each SSM define which SSM is connected to its downstream connectors.
        # For example: 
        # self.ssm_connections = [
        #      SystemSyncModuleConnection(ssm1_chassis=1, ssm1_downstream_connector_number=0, ssm2_chassis=2)]
        self.ssm_connections = []
        
        # Assign triggers to HVI object to be used for HVI-managed synchronization, data sharing, etc
        # NOTE: In a multi-chassis setup ALL the PXI lines listed below need to be shared among each M9031 board pair by means of SMB cable connections
        self.pxi_sync_trigger_resources = [
            kthvi.TriggerResourceId.PXI_TRIGGER0,
            kthvi.TriggerResourceId.PXI_TRIGGER1,
            kthvi.TriggerResourceId.PXI_TRIGGER2,
            kthvi.TriggerResourceId.PXI_TRIGGER3]

######################################
# Main Program
######################################

def main():
    ###################################################
    # Open PXI modules, load PathWave FPGA bitfiles
    ###################################################
    # Load configuration
    config = ApplicationConfig()

    # Checks flags defining if the code executes on HW/Simulation, SD1
    if config.hardware_simulated:
        print("Code running in Simulation Mode")

    # Open instruments
    module_dict = open_modules(config)

    # Load bitfiles generated with PathWave FPGA
    if not config.hardware_simulated:
        load_fpga_bitfiles(module_dict, config)

    ######################################
    # System definition
    ######################################
    my_system = kthvi.SystemDefinition("MySystem") 
    # Define your system, HW platform and HVI resources
    define_hvi_resources(my_system, module_dict, config)
    # Output to console the details of the system definition
    output_sys_def(my_system, config.module_descriptors)
    
    ######################################
    # Program HVI sequences
    ######################################
    print("-----------------------")
    print("Program HVI Sequences")
    print("-----------------------")
    print("Initializing the defined system...")
    # Create sequencer object
    sequencer = kthvi.Sequencer("MySequencer", my_system)
    print("Programming the HVI sequences...")
    # Program HVI sequences using the sequencer object
    program_fpga_interaction_sequence(sequencer, config)
    # Export the programmed sequence to text
    export_hvi_seqeunces(sequencer)

    ########################################
    # Compile, Load to HW, Run HVI Sequence
    ########################################
    print("-------------------")
    print("Execute HVI")
    print("-------------------")
    print("Compiling HVI...")
    # Compile HVI sequences
    hvi = sequencer.compile()
    print(hvi.compile_status.to_string())
    print("HVI Compiled")
    print("This HVI needs to reserve {} PXI trigger resources to execute".format(len(hvi.compile_status.sync_resources)))
    
    # Load HVI to HW: load sequences, configure actions/triggers/events, lock resources, etc.
    hvi.load_to_hw()    
    print("HVI Loaded to HW")

    # Execute HVI in non-blocking mode 
    # This mode allows SW execution to interact with HVI execution
    hvi.run(hvi.no_wait)
    print("HVI Running...")

    # User-controlled HVI execution
    if config.hardware_simulated:
        print("Simulation completed successfully")
    else:
        interact_with_hvi(hvi, module_dict, config)

    # Release HW resources once HVI execution is completed
    hvi.release_hw()
    print("Releasing HW...")

    # Close all modules at the end of the execution
    for engine_name in module_dict:
        module_dict[engine_name].instrument.close()
    print("Modules closed\n")


########################################################
# 2nd Level: Functions to Define, Program, Execute HVI
#######################################################

def load_fpga_bitfiles(module_dict, config):
    """
    Loads the bitstream to the sandbox of each PXI modules and displays a summary of the FPGA sandbox resources
    """
    # Leader module, follower module
    leader_module = module_dict[config.leader_engine].instrument
    follower_module = module_dict[config.follower_engine].instrument
    
    # Load custom FPGA FW designed using Keysight Pathwave FPGA
    error = leader_module.FPGAload(config.leader_project_file)
    if error < 0:
        raise Exception("leader_module.FPGAload() error! Error code: {} Error message: {}".format(error, keysightSD1.SD_Error.getErrorMessage(error)))
    error = follower_module.FPGAload(config.follower_project_file)
    if error < 0:
        raise Exception("follower_module.FPGAload() error! Error code: {} Error message: {}".format(error, keysightSD1.SD_Error.getErrorMessage(error)))

    #Get leader module's registers info
    print("Registers contained in Leader Module:")
    registers = leader_module.FPGAgetSandBoxRegisters(config.num_leader_regs)

    #Print the register properties in register list
    for register in registers:
        print("Register name: {}".format(register.Name))
        print("Register size [bytes]: {}".format(register.Length))
        print("Register address offset [bytes]: {}".format(register.Address))
        print("Register access: {}".format(register.AccessType))
    print()

    #Get follower module's registers info
    print("Registers contained in Follower Module:")
    registers = follower_module.FPGAgetSandBoxRegisters(config.num_follower_regs)

    #Print the register properties in register list
    for register in registers:
        print("Register name: {}".format(register.Name))
        print("Register size [bytes]: {}".format(register.Length))
        print("Register address offset [bytes]: {}".format(register.Address))
        print("Register access: {}".format(register.AccessType))
    print()


def define_hvi_resources(sys_def, module_dict, config):
    """
    Configures all the necessary resources for the HVI application to execute: HW platform, engines, actions, triggers, etc. 
    """
    # Define HW platform: chassis, interconnections, PXI trigger resources, synchronization, HVI clocks
    define_hw_platform(sys_def, config)
    
    # Define all the HVI engines to be included in the HVI
    define_hvi_engines(sys_def, module_dict)
    
    # Define FPGA actions, events and other configurations
    define_fpga_resources(sys_def, module_dict, config)


def define_hw_platform(sys_def, config):
    """
     Define HW platform: chassis, interconnections, PXI trigger resources, synchronization, HVI clocks 
    """
    # Add chassis resources 
    # For multi-chassis setup details see programming example documentation
    for chassis_number in config.chassis_list:
        if config.hardware_simulated:
            sys_def.chassis.add_with_options(chassis_number, 'Simulate=True,DriverSetup=Model=GenericPxieChassis')
        else:
            sys_def.chassis.add(chassis_number)

    # Define System Sync Modules (SSMs)
    if config.system_sync_modules_descriptors:
        interconnects = sys_def.interconnects
        ssm_list = []
        for descriptor in config.system_sync_modules_descriptors:
            if config.hardware_simulated:
                ssm = interconnects.add_sync_module(descriptor.resource_id, config.ssm_simulation_options)
            else:
                ssm = interconnects.add_sync_module(descriptor.resource_id, descriptor.options)
            # Append SSM object
            ssm_list.append(ssm)
        # Define connections between SSMs
        if config.ssm_connections:
            for connection in config.ssm_connections:
                connector_number = connection.ssm1_downstream_connector_number
                for ssm in ssm_list:
                    if ssm.chassis == connection.ssm1_chassis:
                        ssm1 = ssm
                    if ssm.chassis == connection.ssm2_chassis:
                        ssm2 = ssm
                # Implement each user-defined connection
                try:
                    # Set connection. SSMs have always one upstream port
                    ssm1.connectivity.systemsync_downstream[connector_number].set_connection(ssm2.connectivity.systemsync_upstream[0])
                except:
                    exit("Exception! Please check the valued defined for SyncModule resource ids, chassis numbers and connections")

    # Assign the defined PXI trigger resources
    sys_def.sync_resources = config.pxi_sync_trigger_resources
    
    # Assign clock frequencies that are outside the set of the clock frequencies of each HVI engine
    # Use the code line below if you want the application to be in sync with the 10 MHz clock
    sys_def.non_hvi_core_clocks = [10e6]


def define_hvi_engines(sys_def, module_dict):
    """
    Define all the HVI engines to be included in the HVI
    """
    # For each instrument to be used in the HVI application add its HVI Engine to the HVI Engine Collection 
    for engine_name, module in zip(module_dict.keys(), module_dict.values()):
        sys_def.engines.add(module.instrument.hvi.engines.main_engine, engine_name)


def define_fpga_resources(sys_def, module_dict, config):
    """ 
    Define FPGA actions, events and other configurations 
    """
    # Leader module, follower module
    leader_module = module_dict[config.leader_engine].instrument
    follower_module = module_dict[config.follower_engine].instrument

    # Events: add FpgaUserEvent4 to the list of events of the leader engine
    fpga_user_event4 = leader_module.hvi.events.fpga_user_4
    sys_def.engines[config.leader_engine].events.add(fpga_user_event4, config.hvi_user_event_4)

    # Actions: add FpgaUserAction4 to the list of actions of the leader engine
    fpga_user_action4 = leader_module.hvi.actions.fpga_user_4
    sys_def.engines[config.leader_engine].actions.add(fpga_user_action4, config.hvi_user_action_4)

    # Get engine sandbox
    sandbox_name = config.M3xxxA_sandbox
    leader_sandbox = sys_def.engines[config.leader_engine].fpga_sandboxes[sandbox_name]
    follower_sandbox = sys_def.engines[config.follower_engine].fpga_sandboxes[sandbox_name]
    # Load to the sandboxes .k7z project created using Pathwave FPGA
    # This operation is necessary for HVI to list all the FPGA blocks contrined in the designed FPGA FW
    leader_sandbox.load_from_k7z(config.leader_project_file)
    follower_sandbox.load_from_k7z(config.follower_project_file)
    
    # Enable PXI lines to be written from the FPGA sandbox of leader engine only using FPGATriggerOutConfig()
    # NOTE: Only one PXI module per segment shall be allowed to write backplane PXI lines. It would cause conflicts and misbehavior to configure the PXI lines for the follower engine also 
    leader_module.FPGATriggerConfig(externalSource=keysightSD1.SD_TriggerExternalSources.TRIGGER_PXI4, direction =keysightSD1.SD_FpgaTriggerDirection.INOUT, polarity= keysightSD1.SD_TriggerPolarity.ACTIVE_LOW, syncMode = keysightSD1.SD_SyncModes.SYNC_NONE, delay5Tclk=0)
    leader_module.FPGATriggerConfig(externalSource=keysightSD1.SD_TriggerExternalSources.TRIGGER_PXI5, direction =keysightSD1.SD_FpgaTriggerDirection.INOUT, polarity= keysightSD1.SD_TriggerPolarity.ACTIVE_LOW, syncMode = keysightSD1.SD_SyncModes.SYNC_NONE, delay5Tclk=0)
    leader_module.FPGATriggerConfig(externalSource=keysightSD1.SD_TriggerExternalSources.TRIGGER_PXI6, direction =keysightSD1.SD_FpgaTriggerDirection.INOUT, polarity= keysightSD1.SD_TriggerPolarity.ACTIVE_LOW, syncMode = keysightSD1.SD_SyncModes.SYNC_NONE, delay5Tclk=0)
    leader_module.FPGATriggerConfig(externalSource=keysightSD1.SD_TriggerExternalSources.TRIGGER_PXI7, direction =keysightSD1.SD_FpgaTriggerDirection.INOUT, polarity= keysightSD1.SD_TriggerPolarity.ACTIVE_LOW, syncMode = keysightSD1.SD_SyncModes.SYNC_NONE, delay5Tclk=0)

    follower_module.FPGATriggerConfig(externalSource=keysightSD1.SD_TriggerExternalSources.TRIGGER_PXI4, direction =keysightSD1.SD_FpgaTriggerDirection.IN, polarity= keysightSD1.SD_TriggerPolarity.ACTIVE_LOW, syncMode = keysightSD1.SD_SyncModes.SYNC_NONE, delay5Tclk=0)
    follower_module.FPGATriggerConfig(externalSource=keysightSD1.SD_TriggerExternalSources.TRIGGER_PXI5, direction =keysightSD1.SD_FpgaTriggerDirection.IN, polarity= keysightSD1.SD_TriggerPolarity.ACTIVE_LOW, syncMode = keysightSD1.SD_SyncModes.SYNC_NONE, delay5Tclk=0)
    follower_module.FPGATriggerConfig(externalSource=keysightSD1.SD_TriggerExternalSources.TRIGGER_PXI6, direction =keysightSD1.SD_FpgaTriggerDirection.IN, polarity= keysightSD1.SD_TriggerPolarity.ACTIVE_LOW, syncMode = keysightSD1.SD_SyncModes.SYNC_NONE, delay5Tclk=0)
    follower_module.FPGATriggerConfig(externalSource=keysightSD1.SD_TriggerExternalSources.TRIGGER_PXI7, direction =keysightSD1.SD_FpgaTriggerDirection.IN, polarity= keysightSD1.SD_TriggerPolarity.ACTIVE_LOW, syncMode = keysightSD1.SD_SyncModes.SYNC_NONE, delay5Tclk=0)


def output_sys_def(sys_def, descriptors):
    """
    Outputs the details of each instrument and engine included in the system definition.
    """
    print("-------------------")
    print("System Definition")
    print("-------------------")
    # Output KS2201A HVI Core Version
    hvi_core_version = kthvi.SystemDefinition.hvi_core_version
    print("The PathWave Test Sync Executive API installed on your system has an HVI core version " + hvi_core_version.to_string() + "\n")
    # Output info about the HW instruments
    print("HW Instruments:")
    for descriptor in descriptors:
        engine = sys_def.engines[descriptor.engine_name]
        print("- Model: {} in chassis: {}, slot: {}, HVI Engine Name: {}, HVI core version: ".format(descriptor.model_number, descriptor.chassis_number, descriptor.slot_number, descriptor.engine_name) + 
        engine.software_version.to_string() + ", HVI Engine FPGA IP version: " + engine.firmware_version.to_string())
    # Output info about the Sync Modules
    if sys_def.interconnects:
        print("System Sync Modules:")
        for index in range(sys_def.interconnects.count):
            sync_module = sys_def.interconnects[index]
            len_upstream = len(sync_module.connectivity.systemsync_upstream)
            len_downstream = len(sync_module.connectivity.systemsync_downstream)
            print("- System Sync Module in chassis: {}, slot: {} with {} Upstream Ports and {} Downstream Ports".format(sync_module.chassis, sync_module.slot, len_upstream, len_downstream))


def program_fpga_interaction_sequence(sequencer, config):
    """
    This method programs the HVI sequence of this application.
    Different HVI statements are encapsulated as much as possible in separated SW methods to help users visualize
    the programmed HVI sequences.
    The programming example documentation on www.keysight.com contains an HVI diagram that graphically represents the programmed HVI sequence.
    """
    # Define registers within the scope of the outmost sync sequence
    define_registers(sequencer, config)

    # Define Sync While condition
    hvi_quit = sequencer.sync_sequence.scopes[config.leader_engine].registers[config.hvi_quit]
    sync_while_condition = kthvi.Condition.register_comparison(hvi_quit, kthvi.ComparisonOperator.NOT_EQUAL_TO, 1)
    # Add Sync While statement
    sync_while = sequencer.sync_sequence.add_sync_while("User-controlled sync loop", 90, sync_while_condition)
    # Program sync loop 
    program_sync_loop(sync_while.sync_sequence, config)


def define_registers(sequencer, config):
    """
    Defines all registers for each HVI engine in the scope af the global sync sequence 
    """
    # Define registers for leader engine
    leader_engine_register_collection = sequencer.sync_sequence.scopes[config.leader_engine].registers
    hvi_quit = leader_engine_register_collection.add(config.hvi_quit, kthvi.RegisterSize.SHORT)
    hvi_quit.initial_value = 0
    action4_cnt = leader_engine_register_collection.add(config.action4_cnt, kthvi.RegisterSize.SHORT)
    action4_cnt.initial_value = 0
    counter_reg = leader_engine_register_collection.add(config.counter_reg, kthvi.RegisterSize.SHORT)
    counter_reg.initial_value = 0
    mem_map = leader_engine_register_collection.add(config.mem_map, kthvi.RegisterSize.SHORT)
    mem_map.initial_value = 0
    mem_map_counter = leader_engine_register_collection.add(config.mem_map_counter, kthvi.RegisterSize.SHORT)
    mem_map_counter.initial_value = 1000
    pxi_values = leader_engine_register_collection.add(config.pxi_values, kthvi.RegisterSize.SHORT)
    pxi_values.initial_value = 0
    
    # Define registers for leader engine
    follower_engine_register_collection = sequencer.sync_sequence.scopes[config.follower_engine].registers
    follower_counter_reg = follower_engine_register_collection.add(config.follower_counter_reg, kthvi.RegisterSize.SHORT)
    follower_counter_reg.initial_value = 0
    follower_pxi_values = follower_engine_register_collection.add(config.follower_pxi_values, kthvi.RegisterSize.SHORT)
    follower_pxi_values.initial_value = 0


def program_sync_loop(sync_sequence, config):
    """
    Adds and programs the outmost Sync While statement of the HVI Sync Sequence
    """
    # Add 1st Sync Multi-Sequence Block to the Sync While sequence
    sync_block_1 = sync_sequence.add_sync_multi_sequence_block("FPGA Read/Write Operations", 260)
    # Add and program 1st Sync Multi-Sequence Block
    program_fpga_operations(sync_block_1, config)

    # In this Sync Multi-Seq. Block both leader and follower modules read into an HVI register the PXI lines containing the action4 counter value.
    # NOTE: It is critical for the reading operation to be successful to allow enough delay from the previous SMSB where the action4 counter
    # value was written to PXI lines. This is done using the pxi_propagation_delay Python variable. 
    # The necessary pxi_propagation_delay value depends on the module position (segments, chassis).
    # When both leader and follower modules are placed in the same segment, or adjacent segments, pxi_propagation_delay = 100 ns shall be enough.
    # The more the follower and leader module are far away, the more the pxi_propagation_delay value shall be increased.
    pxi_propagation_delay = 300 #[ns]
    # Add 2nd Sync Multi-Sequence Block (SMSB) to the Sync While sequence
    sync_block_2 = sync_sequence.add_sync_multi_sequence_block("Wait for HVI_UserEvent4 and Execute HVI_UserAction4", pxi_propagation_delay)
    # Add and program 2nd Sync Multi-Sequence Block
    program_wait_event_block(sync_block_2, config)

    # Add a 3rd Sync Multi-Sequence Block to the Sync While sequence
    sync_block_3 = sync_sequence.add_sync_multi_sequence_block("Increment counter registers", 10)
    # Add and program 3rd Sync Multi-Sequence Block
    program_increment_counters(sync_block_3, config)


def program_fpga_operations(sync_block, config):
    """
    Programs a Sync Multi-Sequence Block (SMSB) statement to execute FPGA Read/Write Operations
    """
    # Local sequence objects within the SMSB
    leader_sequence = sync_block.sequences[config.leader_engine]

    #Previously defined registers
    action4_cnt = leader_sequence.scope.registers[config.action4_cnt]
    mem_map_counter = leader_sequence.scope.registers[config.mem_map_counter]
    mem_map = leader_sequence.scope.registers[config.mem_map]
    #Previously defined FPGA resources
    fpga_reg_action4_cnt = leader_sequence.engine.fpga_sandboxes[config.M3xxxA_sandbox].fpga_registers[config.reg_action4_cnt]
    fpga_reg_pxi_out = leader_sequence.engine.fpga_sandboxes[config.M3xxxA_sandbox].fpga_registers[config.reg_pxi_out]
    fpga_memory_map = leader_sequence.engine.fpga_sandboxes[config.M3xxxA_sandbox].fpga_memory_maps[config.memory_map]
    
    # Read FPGA Register Register_Bank_HviAction4Cnt
    readFpgaReg0 = leader_sequence.add_instruction("Read FPGA Register_Bank_HviAction4Cnt", 10, leader_sequence.instruction_set.fpga_register_read.id)
    readFpgaReg0.set_parameter(leader_sequence.instruction_set.fpga_register_read.destination.id, action4_cnt)
    readFpgaReg0.set_parameter(leader_sequence.instruction_set.fpga_register_read.fpga_register.id, fpga_reg_action4_cnt)

    # Write FPGA Register Register_Bank_HviPxiTrigOut
    # Register_Bank_HviPxiTrigOut is connected to PXI lines Outputs. The value written to the FPGA register will be written to PXI lines
    # NOTE: Please allow at least 60 ns between these instructions to ensure the HVI register action4_cnt is updated before writing its content to PXI lines
    writeFpgaReg0 = leader_sequence.add_instruction("Write FPGA Register_Bank_HviPxiTrigOut", 60, leader_sequence.instruction_set.fpga_register_write.id)
    writeFpgaReg0.set_parameter(leader_sequence.instruction_set.fpga_register_write.fpga_register.id, fpga_reg_pxi_out)
    writeFpgaReg0.set_parameter(leader_sequence.instruction_set.fpga_register_write.value.id, action4_cnt)

    # Write Memory Map
    # At each iteration a different value is written to the memory map
    writeMemoryMap = leader_sequence.add_instruction("Write FPGA Memory Map", 10, leader_sequence.instruction_set.fpga_array_write.id)
    writeMemoryMap.set_parameter(leader_sequence.instruction_set.fpga_array_write.fpga_memory_map.id, fpga_memory_map)
    writeMemoryMap.set_parameter(leader_sequence.instruction_set.fpga_array_write.value.id, mem_map_counter)
    writeMemoryMap.set_parameter(leader_sequence.instruction_set.fpga_array_write.fpga_memory_map_offset.id, 0)

    # Read Memory Map
    # Reads the value that was written to the block RAM connected to the memory map
    # NOTE: Please allow at least 30 ns between these instructions to ensure data is written correctly through the memory map before you read it back
    readMemoryMap = leader_sequence.add_instruction("Read FPGA Memory Map", 30, leader_sequence.instruction_set.fpga_array_read.id)
    readMemoryMap.set_parameter(leader_sequence.instruction_set.fpga_array_read.fpga_memory_map.id, fpga_memory_map)
    readMemoryMap.set_parameter(leader_sequence.instruction_set.fpga_array_read.destination.id, mem_map)
    readMemoryMap.set_parameter(leader_sequence.instruction_set.fpga_array_read.fpga_memory_map_offset.id, 0)

    return


def program_wait_event_block(sync_block, config):
    """
    Programs a Sync Multi-Sequence Block (SMSB) statement to wait for a sandbox event and execute a sandbox action
    """
    # Local sequence objects within the SMSB
    leader_sequence = sync_block.sequences[config.leader_engine]
    follower_sequence = sync_block.sequences[config.follower_engine]

    # Previously defined registers
    pxi_values = leader_sequence.scope.registers[config.pxi_values]
    follower_pxi_values = follower_sequence.scope.registers[config.follower_pxi_values]
    # Previously defined FPGA resources
    fpga_reg_pxi_in = leader_sequence.engine.fpga_sandboxes[config.M3xxxA_sandbox].fpga_registers[config.reg_pxi_in]
    fpga_follower_reg_pxi_in = follower_sequence.engine.fpga_sandboxes[config.M3xxxA_sandbox].fpga_registers[config.follower_reg_pxi_in]

    # Both Leader and Follower modules read PXI liness logical values from FPGA Register Register_Bank_HviPxiTrigIn
    # In the sandbox of both leader and follower modules sandbox FPGA Register Register_Bank_HviPxiTrigIn is connected to PXI lines inputs
    readFpgaReg0 = leader_sequence.add_instruction("Read Register_Bank_HviPxiTrigIn", 10, leader_sequence.instruction_set.fpga_register_read.id)
    readFpgaReg0.set_parameter(leader_sequence.instruction_set.fpga_register_read.destination.id, pxi_values)
    readFpgaReg0.set_parameter(leader_sequence.instruction_set.fpga_register_read.fpga_register.id, fpga_reg_pxi_in)
    # Read PXI lines from FPGA Register Register_Bank_HviPxiTrigIn in follower instrument 
    readFollowerFpgaReg2 = follower_sequence.add_instruction("Read Register_Bank_HviPxiTrigIn", 10, follower_sequence.instruction_set.fpga_register_read.id)
    readFollowerFpgaReg2.set_parameter(follower_sequence.instruction_set.fpga_register_read.destination.id, follower_pxi_values)
    readFollowerFpgaReg2.set_parameter(follower_sequence.instruction_set.fpga_register_read.fpga_register.id, fpga_follower_reg_pxi_in)

    # Wait for FPGA_User_Event4
    # Define the condition for the wait statement
    wait_condition = kthvi.Condition.event(leader_sequence.engine.events[config.hvi_user_event_4])
    # Add wait statement
    waitEvent = leader_sequence.add_wait("Wait for FPGA_User_Event4", 10, wait_condition)
    waitEvent.set_mode(kthvi.WaitMode.TRANSITION, kthvi.SyncMode.IMMEDIATE)

    # Action execute instruction: execute action 4
    instAction4 = leader_sequence.add_instruction("Execute Action 4", 20, leader_sequence.instruction_set.action_execute.id)
    instAction4.set_parameter(leader_sequence.instruction_set.action_execute.action.id, leader_sequence.engine.actions[config.hvi_user_action_4])

    return


def program_increment_counters(sync_block, config):
    """
    Programs a Sync Multi-Sequence Block (SMSB) statement to increment register counters
    """
    # Local sequence objects within the SMSB
    leader_sequence = sync_block.sequences[config.leader_engine]
    follower_sequence = sync_block.sequences[config.follower_engine]

    #Previously defined registers
    counter_reg = leader_sequence.scope.registers[config.counter_reg]
    mem_map_counter = leader_sequence.scope.registers[config.mem_map_counter]
    follower_counter_reg = follower_sequence.scope.registers[config.follower_counter_reg]

    # Increment memory map counter
    instr = leader_sequence.add_instruction("Increment memory map counter", 10, leader_sequence.instruction_set.add.id)
    instr.set_parameter(leader_sequence.instruction_set.add.left_operand.id, mem_map_counter)
    instr.set_parameter(leader_sequence.instruction_set.add.right_operand.id, 1)
    instr.set_parameter(leader_sequence.instruction_set.add.destination.id, mem_map_counter)

    # Increment counter register
    instr = leader_sequence.add_instruction("Increment counter register", 10, leader_sequence.instruction_set.add.id)
    instr.set_parameter(leader_sequence.instruction_set.add.left_operand.id, counter_reg)
    instr.set_parameter(leader_sequence.instruction_set.add.right_operand.id, 1)
    instr.set_parameter(leader_sequence.instruction_set.add.destination.id, counter_reg)

    # Increment follower counter register
    instr = follower_sequence.add_instruction("Increment Follower counter register", 10, follower_sequence.instruction_set.add.id)
    instr.set_parameter(follower_sequence.instruction_set.add.left_operand.id, follower_counter_reg)
    instr.set_parameter(follower_sequence.instruction_set.add.right_operand.id, 1)
    instr.set_parameter(follower_sequence.instruction_set.add.destination.id, follower_counter_reg)


def export_hvi_seqeunces(sequencer):
    """
    Exports the programmed HVI sequences to text format
    """
    # Generate HVI sequence description text
    output = sequencer.sync_sequence.to_string(kthvi.OutputFormat.DEBUG)
    print("Programmed HVI sequences exported to file")
    
    # Write file
    text_file = open("../TextOutput/HviPwFpgaIntegration.txt", "w")
    text_file.write(output)
    text_file.close()


def interact_with_hvi(hvi, module_dict, config):
    """
    This function allows users to interact with the HVI execution from SW execution
    """ 
    # Leader engine registers
    leader_engine_scope = hvi.sync_sequence.scopes[config.leader_engine]
    action4_cnt = leader_engine_scope.registers[config.action4_cnt]
    mem_map = leader_engine_scope.registers[config.mem_map]
    mem_map_counter_reg = leader_engine_scope.registers[config.mem_map_counter]
    pxi_values = leader_engine_scope.registers[config.pxi_values]
    counter_reg = leader_engine_scope.registers[config.counter_reg]
    hvi_quit = leader_engine_scope.registers[config.hvi_quit]
    # Follower engine registers
    follower_engine_scope = hvi.sync_sequence.scopes[config.follower_engine]
    follower_pxi_values = follower_engine_scope.registers[config.follower_pxi_values]
    follower_counter_reg = follower_engine_scope.registers[config.follower_counter_reg]

    # User controlled loop
    leader_sandbox = hvi.sync_sequence.engines[config.leader_engine].fpga_sandboxes[config.M3xxxA_sandbox]
    while True:
        print("N. of User Actions counted at leader module: action4_cnt = {}".format(action4_cnt.read()))
        print("Value written to the FPGA Memory Map: mem_map = {}".format(mem_map.read()))
        print("Value read by Leader Module from FPGA PXI inputs: pxi_values = {}".format(pxi_values.read()))
        print("Value read by Follower Module from FPGA PXI inputs: follower_pxi_values = {}".format(follower_pxi_values.read()))
        print("Counter value: counter_reg = {}".format(counter_reg.read()))
        print("Follower counter value: follower_counter_reg = {}".format(follower_counter_reg.read()))
        print("Mem. Map counter value: mem_map_counter_reg = {}".format(mem_map_counter_reg.read()))

        # Uncomment to print debug FPGA register values
        # debug_sandbox(module_dict, config)

        print("Press enter to trigger a User Event and execute a User Action, press q to exit")
        key = input()

        # Generate an event or quit
        if key == 'q':
            hvi_quit.write(1)
            time.sleep(.001)
            trigger_fpga_user_event(leader_sandbox, config)
            break
        else:
            trigger_fpga_user_event(leader_sandbox, config)


def trigger_fpga_user_event(sandbox, config):
    """
    Triggers an FPGA_Event4 from the sandbox by writing an HVI sandbox register connected to Event4
    """
    # Write a 1-0 to register "Register_Bank_HviEvent4" within an HVI Port Register Bank
    # Please update the registerName to the actual name of your register if necessary 
    regEvent = sandbox.fpga_registers[config.reg_event4]
    regEvent.write(1)
    regEvent.write(0)


def debug_sandbox(module_dict, config):
    """
    Use this function to read FPGA registers defined for debugging
    """
    # Read register values
    leader_module = module_dict[config.leader_engine].instrument
    for reg_name in config.leader_debug_regs:
        reg_value = leader_module.FPGAgetSandBoxRegister(reg_name).readRegisterInt32()
        print("{} Value: {}".format(reg_name, reg_value))


#################################################################
# 3rd Level: Classes and Functions to Use SD1/M3xxxA Instruments
#################################################################

class ModuleDescriptor:
    "Descriptor for module objects"
    def __init__(self, model_number, chassis_number, slot_number, options, engine_name):
        self.model_number = model_number
        self.chassis_number = chassis_number
        self.slot_number = slot_number
        self.options = options
        self.engine_name = engine_name


class SystemSyncModuleDescriptor:
    """
    Descriptor for System Sync Modules
    """
    def __init__(self, resource_id, options):
        self.resource_id = resource_id
        self.options = options


class SystemSyncModuleConnection:
    """
    Describes the connection between two SyncModule objects ssm1 and ssm2. SSMs have always one upstream port:
    ssm1.connectivity.systemsync_downstream[ssm1_downstream_connector_number].connection = ssm2.connectivity.systemsync_upstream[0]
    """
    def __init__(self, ssm1_chassis, ssm1_downstream_connector_number, ssm2_chassis):
        self.ssm1_chassis = ssm1_chassis
        self.ssm1_downstream_connector_number = ssm1_downstream_connector_number
        self.ssm2_chassis = ssm2_chassis


class Module:
    "Class defining a modular instrument object and its properties"
    def __init__(self, instrument_object, num_channels):
        self.instrument = instrument_object
        self.num_channels = num_channels


def open_modules(config):
    """ 
    Opens and creates all the necessary instrument objects.
    Returns a dictionary of module objects whose keys are the HVI engine names.
    Please check SD1 3.x User Manual for options to open the instrument objects
    """
    # Initialize output variables
    num_modules = 0
    module_dict = {} # dictionary of modules

    # Open SD1 instrument objects
    for descriptor in config.module_descriptors:
        if descriptor.model_number == "M3102A":
            instr_obj = keysightSD1.SD_AIN()
        else:
            instr_obj = keysightSD1.SD_AOU()
        instr_obj_options = descriptor.options + ',simulate=true' if config.hardware_simulated else descriptor.options
        id = instr_obj.openWithOptions(descriptor.model_number, descriptor.chassis_number, descriptor.slot_number, instr_obj_options)
        if id < 0:
            raise Exception("Error opening instrument in chassis: {}, slot: {}! Error code: {} - {}. Exiting...".format(descriptor.chassis_number, descriptor.slot_number, id, keysightSD1.SD_Error.getErrorMessage(id)))
        nCh = instr_obj.getOptions("channels")
        if nCh == "CH2":
            num_channels = 2
        elif nCh == "CH4":
            num_channels = 4
        else:
            raise Exception("PXI module in chassis {}, slot {} returned number of channels = {} which is incorrect. Exiting... ".format(instr_obj.getChassis(), instr_obj.getSlot(), nCh))
        module_dict[descriptor.engine_name] = Module(instr_obj, num_channels)
        num_modules += 1
    
    return module_dict


if __name__ == '__main__':
    main()
